DAT fileName  byte "AS5055Spi130730e", 0
CON
{{ ****** Public Notes ******
  
  Written by: Duane Degn
  For: Kirk Frasure
  March 24, 2012

  This object can read from AustriaMicroSystem's
  AS5055 magnetic encoders in two different modes.

  With both modes the "Start" method must first
  be called.

  The first read method is 4-wire SPI read only mode.
  In this mode a single device is read using the
  "ShiftIn" method.  The "ShiftIn" method returns
  the value read from the encoder, including
  the error and parity bits.  The calling
  object must process this value (by shifting the value
  two bits to the right) to obtain the encoder's position.

  Since the read command is $FFFF, a pin can be saved by
  pulling the MOSI pin on the encoder high and using the
  value "-1" as MOSI pin when starting the object.

  Pulling the MOSI pin high will only work when reading
  from a single device in read only mode.

  The second read method is 4-wire SPI daisy-chain mode.
  After calling the "Start" method, the "Reset" method
  should be called prior to reading from the encoders.

  The "ReadChain" method is used to read the encoder
  data when using daisy-chain mode.  The calling method
  provides a pointer to an array of longs where the
  encoder readings are to be stored.

  As with the single device read the value stored in
  the array by the "ReadChain" method include the
  parity and error bits.

  While the AS5055 devices output data as they
  read data, the duplex nature of the device is
  not used in this object since the read command
  can simulated by setting the MOSI pin high as
  data is read in.
  
}}
{ ****** Private Notes ****** 
  Modified from dwd_nordicSpi110416a.

  120318a Clear tempBits after its value has been written to hub.
  18 Change the way data is clocked in.  Raise clock first, then
  read data.
  21a Start adding daisy-chain capability.
  21b Change some names.  Use a pointer as a
  parameter in ReadChain method.
  23c Clean up code.
  130728a Move filter from parent object to child object. Not implemented yet.
  29a Start moving filter.
  30c Works with a single read.
  30d Remove self modifying code for now.
  
}
CON

  {MAX_ALLOWED_STATIC = 128
  SAMPLES_TO_AVERAGE = 64 '128
  SAMPLES_TO_AVERAGE_MOVING = 16
  MOVING_TRIGGER = 8            ' difference between two concecutive readings to trigger using small buffer to average
  ROLLOVER_ZONE = 1000 }
  
  ' localMovingFlag enumeration
  #0, NOTMOVING_THIS_LOOP, MOVING_THIS_LOOP

  ' displayMode enumeration
  #0, SCROLL_DISPLAY, STATIONARY_DISPLAY

  ' WIRE_MODE enumeration
  #0, THREE_WIRE_MODE, FOUR_WIRE_MODE

  ' INTERRUPT_MODE enumeration
  #0, NO_INTERRUPT, YES_INTERRUPT
  
  ' Used determine which PASM code is executed.   
  #1, SHIFT_IN, SHIFT_IN_CHAIN_ONCE, INIT_CHAIN_CONTINUOUS, RESET_CMD, CONTINUE_CHAIN_CONTINUOUS 
  
VAR

  long cog, command
  long valueIn, globalPtr, chainLength 
  long parameterPtrHub, bufferPtrHub
  ' Variables duplicated in PASM of child object.
  long bufferHead
  long bufferHeadM
  ' Variables used by child object. Keep the varaiables below together and in order.
  long samplesToAverageStatic, samplesToAverageMoving, movingThreshold
  long bufferSizeStatic, bufferSizeMoving, wireMode, interruptMode
  
  {long bufferTotal[CHAINED_DEVICES], bufferTotalMov[CHAINED_DEVICES]
  long average[CHAINED_DEVICES], averageMov[CHAINED_DEVICES]
  long rolloverTotal[CHAINED_DEVICES], rolloverTotalMov[CHAINED_DEVICES]
  long rolloverCount[CHAINED_DEVICES], rolloverCountMov[CHAINED_DEVICES]
  long notMovingCount[CHAINED_DEVICES]
 
  long averagingBuffer[BUFFER_SIZE_STATIC], averagingBufferMov[BUFFER_SIZE_MOVING]
  long rolloverBuffer[BUFFER_SIZE_STATIC], rolloverBufferMov[BUFFER_SIZE_MOVING]
  ' Keep the varaiables above together and in order.
  byte movingState[CHAINED_DEVICES]
  byte rolloverState[CHAINED_DEVICES], rolloverStateMov[CHAINED_DEVICES]
  }
    
PUB Start(_mosi, _miso, _clock, _ss, _int, parameterPtr, bufferPtr) : okay
'' Start SPI Engine - starts a cog
'' returns false if no cog available

  Stop    

  if _mosi == -1  ' Used to save IO pin.  MOSI is pulled high externally.
    mosiMask := 0 ' ReadChain method can't be used in read only mode.
  else
    mosiMask := 1 << _mosi

  if _int > 31 or _int < 0  ' If interrupt pin isn't a valid pin number, then assume
    intMask := 0 ' it's a delay amount (in clock cycles).
    readPause := _int
  else
    intMask := 1 << _int

  misoMask := 1 << _miso  
  clockMask := 1 << _clock
  ssMask := 1 << _ss
    
  globalPtrLoc := @globalPtr
  parameterPtrCog := parameterPtr
  bufferPtrCog := bufferPtr
  InitSpinFilter(parameterPtr, bufferPtr)
  
  inWordPtr := @valueIn
  commandPtr := @command
  chainLengthPtr := @chainLength
      
  okay := cog := cognew(@enter, @command) + 1
  ' The cog will take a few ms to start.

PUB InitSpinFilter(parameterPtr, bufferPtr)
'' Probably wont use this method.

  parameterPtrHub := parameterPtr
  bufferPtrHub := bufferPtr
  longmove(@samplesToAverageStatic, parameterPtrHub, 7)
  'if interruptMode == NO_INTERRUPT
    
PUB Stop
'' Stop SPI Engine - frees a cog
    if cog
       cogstop(cog~ - 1)
    command~
    
PUB ShiftIn        
'' Returns a word read in by SPI
                                          
  setcommand(SHIFT_IN)
   
  Result := valueIn

PUB ReadChain(localPtr, _chainLength)        
'' Returns an address to array of values.

  globalPtr := localPtr
  chainLength := _chainLength   
                                                                              
  setcommand(SHIFT_IN_CHAIN_ONCE)
  

{PUB ReadFiltered'(localPtr, _chainLength)        
'' Returns an address to array of values.

  globalPtr := localPtr
  chainLength := _chainLength   
                                                                              
  setcommand(SHIFT_IN_CHAIN_ONCE)
  
 }
PUB ReadContinuous(localPtr, _chainLength)        
'' Returns an address to array of values.

  globalPtr := localPtr
  chainLength := _chainLength   
                                                                              
  command := INIT_CHAIN_CONTINUOUS
  repeat while command == INIT_CHAIN_CONTINUOUS
  

PUB Reset(_chainLength)        
'' Reset encoders

  chainLength := _chainLength     
  setcommand(RESET_CMD)

PRI Setcommand(cmd)
    command := cmd              ''write command
    repeat while command        ''wait for command to be cleared, signifying receipt

PUB GetFileName
'' This method lets a parent object read the fileName
'' of this object.
'' With many projects, I like to be able to see
'' What code is currently loaded in the Propeller.

  result := @fileName

{PRI ProcessAngle(localIndex) | parity, error, localMovingFlag
'' Parity bit is not tested or used.
'' localIndex is the ID of the device to process.
'' The device connected to the Prop's MISO pin is
'' device #0.
'' This method performs all the software filtering of the data.
       
  parity := encoderAngle[localIndex] & %1
  encoderAngle[localIndex] >>= 1
  error := encoderAngle[localIndex] & %1
  
  if error == 1 ' If there is an error, display the binary value of the data.
    Debug.str(string(13, " encoder #"))
    Debug.dec(localIndex)
    Debug.str(string(" returned an error. %")) 
    Debug.bin(encoderAngle[localIndex], 31)
    Debug.bin(parity, 1)
    result := 1                 ' Inform calling method of error.
  encoderAngle[localIndex] >>= 1
  encoderAngle[localIndex] &= _TwelveBits

  ' Compute averages, check to see if value is near rollover point and
  ' check to see if the encoder is moving.
  
  bufferTotal[localIndex] -= averagingBuffer[(localIndex * _SamplesToAverage) + bufferHead]
  averagingBuffer[(localIndex * _SamplesToAverage) + bufferHead] := encoderAngle[localIndex]
  bufferTotal[localIndex] += averagingBuffer[(localIndex * _SamplesToAverage) + bufferHead]
  average[localIndex] := bufferTotal[localIndex] / _SamplesToAverage

  bufferTotalM[localIndex] -= averagingBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM]
  averagingBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM] := encoderAngle[localIndex]
  bufferTotalM[localIndex] += averagingBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM]
  averageM[localIndex] := bufferTotalM[localIndex] / _SamplesToAverageMoving

   
  if encoderAngle[localIndex] < _RolloverZone or encoderAngle[localIndex] > _TwelveBits - _RolloverZone
    rolloverTotal[localIndex] -= rolloverBuffer[(localIndex * _SamplesToAverage) + bufferHead]
    rolloverTotalM[localIndex] -= rolloverBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM]
    if encoderAngle[localIndex] < _RolloverZone
      rolloverBuffer[(localIndex * _SamplesToAverage) + bufferHead] := encoderAngle[localIndex] + _TwelveBitsPlusOne
      rolloverBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM] := encoderAngle[localIndex] + _TwelveBitsPlusOne
    else
      rolloverBuffer[(localIndex * _SamplesToAverage) + bufferHead] := encoderAngle[localIndex]
      rolloverBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM] := encoderAngle[localIndex]
    rolloverTotal[localIndex] += rolloverBuffer[(localIndex * _SamplesToAverage) + bufferHead]
    rolloverTotalM[localIndex] += rolloverBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM]  
    if rolloverState[localIndex] == _OutOfZoneEmptyBuffer
      rolloverState[localIndex] := _InZoneFillingBuffer
      rolloverCount[localIndex] := 0
    elseif rolloverState[localIndex] == _InZoneFillingBuffer
      if ++rolloverCount[localIndex] == _SamplesToAverage
        rolloverState[localIndex] := _InZoneUseBuffer
    elseif rolloverState[localIndex] == _InZoneUseBuffer
      average[localIndex] := (rolloverTotal[localIndex]  / _SamplesToAverage) // _TwelveBitsPlusOne
    if rolloverStateM[localIndex] == _OutOfZoneEmptyBuffer
      rolloverStateM[localIndex] := _InZoneFillingBuffer
      rolloverCountM[localIndex] := 0
    elseif rolloverStateM[localIndex] == _InZoneFillingBuffer
      if ++rolloverCountM[localIndex] == _SamplesToAverage
        rolloverStateM[localIndex] := _InZoneUseBuffer
    elseif rolloverStateM[localIndex] == _InZoneUseBuffer
      averageM[localIndex] := (rolloverTotalM[localIndex]  / _SamplesToAverageMoving) // _TwelveBitsPlusOne  
  else ' assume rollover problems have cleared out of buffer by now
    rolloverState[localIndex] := _OutOfZoneEmptyBuffer 
    rolloverStateM[localIndex] := _OutOfZoneEmptyBuffer

  localMovingFlag := _NotMovingThisLoop
   
  if encoderAngle[localIndex] < _MovingTrigger and previousAngle[localIndex] > _TwelveBits - _MovingTrigger
    if (encoderAngle[localIndex] + _TwelveBits) - previousAngle[localIndex] => _MovingTrigger
      localMovingFlag := _MovingThisLoop
  elseif encoderAngle[localIndex] > _TwelveBits - _MovingTrigger and previousAngle[localIndex] < _MovingTrigger
    if (previousAngle[localIndex] + _TwelveBits) -  encoderAngle[localIndex] => _MovingTrigger
      localMovingFlag := _MovingThisLoop
  elseif ||(encoderAngle[localIndex] - previousAngle[localIndex]) => _MovingTrigger  
    localMovingFlag := _MovingThisLoop

  if localMovingFlag == _NotMovingThisLoop
    if movingState[localIndex] == _Moving
      movingState[localIndex] := _NotMovingFillingBuffer
      notMovingCount[localIndex] := 1
      average[localIndex] := averageM[localIndex]
    elseif movingState[localIndex] == _NotMovingFillingBuffer
      if ++notMovingCount[localIndex] => _SamplesToAverage
        movingState[localIndex] := _NotMovingUpdatedBuffer
      else
        average[localIndex] := averageM[localIndex]
  else
    movingState[localIndex] := _Moving
    average[localIndex] := averageM[localIndex]
  
  encoderAngle7[localIndex] := average[localIndex] 
  if average[localIndex] <> previousAngle[localIndex]   '' Comment out this line to display encoder value in each loop.
    if displayMode == _ScrollDisplay 
      Debug.str(string(13, " encoder["))
      Debug.dec(localIndex)
      Debug.str(string("] = "))
      Debug.dec(average[localIndex])
    else        
      DisplayData(localIndex)
    {SevenSegment[localIndex].Char("v")
    if encoderAngle7[localIndex] < 1000
      SevenSegment[localIndex].Char("0")
    if encoderAngle7[localIndex] < 100
      SevenSegment[localIndex].Char("0") 
    if encoderAngle7[localIndex] < 10
      SevenSegment[localIndex].Char("0") 
    SevenSegment[localIndex].Dec(encoderAngle7[localIndex])
    }
    previousAngle[localIndex] := average[localIndex]
 }         
DAT                                             org
'  
'' SPI Engine 
'
enter                                           or      outa, mosiMask        
                                                or      dira, mosiMask              
                                                andn    outa, clockMask        
                                                or      dira, clockMask
                                                or      outa, ssMask             
                                                or      dira, ssMask

                                                'tjz     intMask, #overwriteInterrupt
                                                 
preloop                                         wrlong  zero, commandPtr        ' calling method waits until command is completed  
loop                                            rdlong  receivedCommand, commandPtr ' watch for command
                                                tjz     receivedCommand, #loop
                                                 
                                                mov     jumpMode, receivedCommand              
                                                add     jumpMode, #jumpTable
                                                sub     jumpMode, #1            ' Since zero is used to indicate no command              
                                                jmp     jumpMode                ' indirect jump
                                                 

jumpTable                                       jmp     #shiftInWord_
                                                jmp     #shiftInChain_
                                                jmp     #initChainContinuous
                                                jmp     #reset_
                                                
  ' Used determine which PASM code is executed.   
'  #1, SHIFT_IN, SHIFT_IN_CHAIN_ONCE, INIT_CHAIN_CONTINUOUS, RESET_CMD, CONTINUE_CHAIN_CONTINUOUS 
'------------------------------------------------------------------------------------------------------------------------------

DAT initChainContinuous                         mov     currentJump, countinuousChainJump
                                                wrlong  continue, commandPtr
'------------------------------------------------------------------------------------------------------------------------------

DAT continueContinuous                          rdlong  receivedCommand, commandPtr
                                                cmp     receivedCommand, continue wz
                                if_nz           jmp     #endChainCountinuous
nextRead                                        call    #delayRoutine 'waitpne intMask, intMask ' wait for interrupt pin to go low
                                                             
'------------------------------------------------------------------------------------------------------------------------------

DAT shiftInChain_                               rdlong  readCountdown, chainLengthPtr
                                                or      outa, mosiMask          ' Writes a $FFFF as byte is read 
                                                andn    outa, ssMask            ' start read by pulling chip select low
                                                rdlong  address, globalPtrLoc
                                                                         
wordInLoop                                      mov     tempBits, bits          ' Load number of data bits
                                                mov     tempData, zero          ' Clear tempData
                        
:bit                                            or      outa, clockMask
                                                nop                             ' This might not be needed
                                                test    misoMask, ina     wc    ' Read Data Bit into 'C' flag
                                                rcl     tempData, #1            ' rotate "C" flag into return value
                                                andn    outa, clockMask
                                                djnz    tempBits, #:bit         ' Decrement tempBits ; jump if not Zero
                                                 
                                                shr     tempData, #2            ' Drop last two bits ***
                                                and     tempData, twelveBits    ' ***
                                                wrlong  tempData, address
                                                add     address, #4             ' Increment to next long address
                                                djnz    readCountdown, #wordInLoop
                                                or      outa, ssMask
                                                or      outa, mosiMask
                                                add     readingsCounter, #1
                                                wrlong  readingsCounter, address ' counter located after data
                                                
currentJump                                     jmp     #preloop                ' This line gets modified
               
countinuousChainJump                            jmp     #continueContinuous               
                                                 
endChainCountinuous                             mov     currentJump, singleChainJump
singleChainJump                                 jmp     #preloop                                
'------------------------------------------------------------------------------------------------------------------------------
' Valid data will be ready 20ns after rising clock.
DAT shiftInWord_                                or      outa, mosiMask          '  Writes a $FF as byte is read
                                                mov     readCountdown, #1
                                                mov     address, inWordPtr
                                                jmp     #wordInLoop
                                                 
'------------------------------------------------------------------------------------------------------------------------------

DAT reset_                                      rdlong  readCountdown, chainLengthPtr
                                                andn    outa, mosiMask                                                                                                                    
                                                andn    outa, ssMask         ' start read by pulling chip select low
                                                                         
:wordLoop                                       mov     tempBits, bits              ''     Load number of data bits
                                                mov     tempData, resetCon
                                                'ror     tempData, #bits         ' move high bit into position to shift off
                        
:bit                                            shl     tempData, #1   wc       ' shift off the bit to send 
                                                muxc    outa, mosiMask          ' Set DataBit HIGH or LOW 
                                                or      outa, clockMask
                                                nop
                                                andn    outa, clockMask
                                                djnz    tempBits, #:bit         ' Decrement tempBits ; jump if not Zero
                                                djnz    readCountdown, #:wordLoop
                                                or      outa, ssMask
                                                or      outa, mosiMask          ' leave data out pin high   
                                                jmp     #preloop                ' wait for next command

'------------------------------------------------------------------------------------------------------------------------------
DAT overwriteInterrupt                          mov     nextRead, replaceIntrerruptWithPause
                                                jmp     #preloop
                                                
replaceIntrerruptWithPause                      call    #delayRoutine

delayRoutine                                    mov     nextTime, cnt
                                                add     nextTime, readPause
                                                waitcnt nextTime, #0
delayRoutine_ret                                ret

DAT '-------------------------------Assembly variables -------------------------------

zero                                            long 0                       ' constants
twelveBits                                      long %1111_1111_1111
continue                                        long 5 'CONTINUE_CHAIN_CONTINUOUS
resetCon                                        long %0111_1000_0000_0000_0111_1000_0000_0000 ' reset command twice, only top 16 bits used

                                                        ' temp variables
receivedCommand                                 long 0                            
tempData                                        long 0 ' Used to hold data being read or written
tempBits                                        long 0 ' Remaining # of bits to read or write
readCountdown                                   long 0 ' Remaining words to read or write
address                                         long 0 ' Used to hold return address of first Argument passed

clockMask                                       long 1
misoMask                                        long 1
mosiMask                                        long 1
ssMask                                          long 1
intMask                                         long 1
bits                                            long 16
inWordPtr                                       long 0
outBytePtr                                      long 0
flagPtr                                         long 0
jumpMode                                        long 0
commandPtr                                      long 0
globalPtrLoc                                    long 0
chainLengthPtr                                  long 0
readPause                                       long 0
parameterPtrCog                                 long 0 
bufferPtrCog                                    long 0
readingsCounter                                 long 0
nextTime                                        res 1
headStatic                                      res 1
headMoving                                      res 1

                                                fit